/*
Copyright (C) 2019 RRe36

All Rights Reserved unless otherwise explicitly stated.


By downloading this you have agreed to the license and terms of use.
These can be found inside the included license-file or here: https://github.com/rre36/glsl_kappa/blob/master/LICENSE

Violating these terms may be penalized with actions according to the Digital Millennium Copyright Act (DMCA), the Information Society Directive and/or similar laws depending on your country.
*/

#include "/lib/common.glsl"
#include "/lib/util/srgb.glsl"

#define INFO 0  //[0]

in vec2 coord;

uniform sampler2D colortex0;
uniform sampler2D colortex3;
uniform sampler2D colortex4;

uniform int frameCounter;

uniform float frameTimeCounter;
uniform float nightVision;

uniform vec2 pixelSize;
uniform vec2 viewSize;

vec3 tonemap_reinhard(vec3 hdr) {
    float luma      = get_luma(hdr);

    float coeff     = 0.9 - nightVision*0.7;

    vec3 col        = hdr/(hdr + coeff);
        col         = mix(hdr/(luma + coeff), col, col);

    return to_srgb(col);
}

struct hable_param {
    float curve;
    float toe;
    float angle;
    float slope;
    float black;
    float range;
    float white;
} hable;

vec3 hable_curve(vec3 x) {
    float A   = hable.curve;
    float B   = hable.toe;
    float C   = hable.slope;
    float D   = hable.angle;
    float E   = hable.black;
    float F   = hable.range;
    return ((x * (A*x + C*B) + D*E) / (x * (A*x + B) + D*F)) - E/F;
}

vec3 tonemap_hable(vec3 hdr) {
    hable.curve        = 0.22;
    hable.toe          = 0.81 + nightVision * 0.15;
    hable.slope        = 0.62;
    hable.angle        = 0.39;
    hable.black        = 0.00;
    hable.range        = 1.70;
    hable.white        = 14.00;

    vec3 white = hable_curve(vec3(hable.white));

    vec3 col    = hable_curve(hdr);
        col     = col * rcp(white);

    return to_srgb(col);
}

vec3 vignette(vec3 color) {
    float fade      = length(coord*2.0-1.0);
        fade        = lin_step(abs(fade), 0.3, 1.8);
        fade        = 1.0-pow3(fade)*0.75;

    return color * fade;
}

float bayer2  (vec2 c) { c = 0.5 * floor(c); return fract(1.5 * fract(c.y) + c.x); }
float bayer4  (vec2 c) { return 0.25 * bayer2 (0.5 * c) + bayer2(c); }
float bayer8  (vec2 c) { return 0.25 * bayer4 (0.5 * c) + bayer2(c); }
float bayer16 (vec2 c) { return 0.25 * bayer8 (0.5 * c) + bayer2(c); }

vec3 dither_image(vec3 color) {
    int bits    = 255;

    vec3 c_dither   = color;
        c_dither   *= bits;
        c_dither   += bayer16(gl_FragCoord.xy);

    return round(c_dither)/bits;
}

/* ------ color grading utilities ------ */

vec3 rgb_luma(vec3 x) {
    x.r *= colorlum_r;
    x.g *= colorlum_g;
    x.b *= colorlum_b;

    return x;
}

/* --- post tonemap color grading --- */
vec3 gammacurve(vec3 x) {
    #if color_preset == 0
    const float gint    = gamma_curve;
    #elif color_preset == 1
    const float gint    = gamma_curve - 0.2;
    #elif color_preset == 2
    const float gint    = gamma_curve + 0.08;
    #endif

    return pow(x, vec3(gint));
}

vec3 vibrance_saturation(vec3 color) {
    #ifdef do_colorgrading
        #if color_preset == 0
        const float vint    = vibrance_int + 0.1;
        const float sint    = saturation_int + 0.02;
        #elif color_preset == 1
        const float vint    = vibrance_int + 0.2;
        const float sint    = saturation_int + 0.18;
        #elif color_preset == 2
        const float vint    = vibrance_int;
        const float sint    = saturation_int + 0.02;
        #endif
    #else
        const float vint    = 1.1;
        const float sint    = 1.02;
    #endif

    float lum   = dot(color, lumacoeff_rec709);
    float mn    = min(min(color.r, color.g), color.b);
    float mx    = max(max(color.r, color.g), color.b);
    float sat   = (1.0 - (mx-mn)) * (1.0-mx) * lum * 5.0;
    vec3 light  = vec3((mn + mx) / 2.0);

    color   = mix(color, mix(light, color, vint), sat);

    color   = mix(color, light, (1.0-light) * (1.0-vint) / 2.0 * abs(vint));

    color   = mix(vec3(lum), color, sint);

    return color;
}

vec3 brightness_contrast(vec3 color) {
    #if color_preset == 0
    const float bint    = brightness_int;
    const float cint    = constrast_int;
    #elif color_preset == 1
    const float bint    = brightness_int;
    const float cint    = constrast_int + 0.13;
    #elif color_preset == 2
    const float bint    = brightness_int + 0.02;
    const float cint    = constrast_int + 0.02;
    #endif

    return (color - 0.5) * cint + 0.5 + bint;
}

vec4 textureBicubic(sampler2D sampler, vec2 coord) {
	vec2 res = textureSize(sampler, 0);

	coord = coord * res - 0.5;

	vec2 f = fract(coord);
	coord -= f;

	vec2 ff = f * f;
	vec4 w0;
	vec4 w1;
	w0.xz = 1 - f; w0.xz *= w0.xz * w0.xz;
	w1.yw = ff * f;
	w1.xz = 3 * w1.yw + 4 - 6 * ff;
	w0.yw = 6 - w1.xz - w1.yw - w0.xz;

	vec4 s = w0 + w1;
	vec4 c = coord.xxyy + vec2(-0.5, 1.5).xyxy + w1 / s;
	c /= res.xxyy;

	vec2 m = s.xz / (s.xz + s.yw);
	return mix(
		mix(texture(sampler, c.yw), texture(sampler, c.xw), m.x),
		mix(texture(sampler, c.yz), texture(sampler, c.xz), m.x),
		m.y);
}

vec4 texture_sharp(sampler2D tex, vec2 uv, const float w) {
    vec2 res    = textureSize(tex, 0);

    vec2 pixelSize = rcp(res);

    float weight = 0.0;

    float corner_weight         = -0.5*w;
    float edge_weight           = 0.46*w;
    const float center_weight   = 1.4;

    vec4 tl     = texture(tex, uv + vec2( 1.0,  1.0)*pixelSize)*corner_weight;    weight += corner_weight;
    vec4 tc     = texture(tex, uv + vec2( 0.0,  1.0)*pixelSize)*edge_weight;    weight += edge_weight;
    vec4 tr     = texture(tex, uv + vec2(-1.0,  1.0)*pixelSize)*corner_weight;    weight += corner_weight;

    vec4 ml     = texture(tex, uv + vec2( 1.0,  0.0)*pixelSize)*edge_weight;    weight += edge_weight;
    vec4 mc     = texture(tex, uv)                  *center_weight;    weight += center_weight;
    vec4 mr     = texture(tex, uv + vec2(-1.0,  0.0)*pixelSize)*edge_weight;    weight += edge_weight;

    vec4 bl     = texture(tex, uv + vec2( 1.0, -1.0)*pixelSize)*corner_weight;    weight += corner_weight;
    vec4 bc     = texture(tex, uv + vec2( 0.0, -1.0)*pixelSize)*edge_weight;    weight += edge_weight;
    vec4 br     = texture(tex, uv + vec2(-1.0, -1.0)*pixelSize)*corner_weight;    weight += corner_weight;

    float norm  = 1.0/weight;

    vec4 col    = (tl+tc+tr+ml+mc+mr+bl+bc+br)*norm;

    if (col.x < 0.0 || col.y < 0.0 || col.z < 0.0) col = mc;

    return col;
}

void main() {
    vec3 scene_hdr  = vec3(0.0);
    
    #ifdef upscale_sharpen
        if (MC_RENDER_QUALITY > 0.9) {
            scene_hdr  = stex(colortex0).rgb;
        } else if (MC_RENDER_QUALITY < 0.6) {
            scene_hdr = texture_sharp(colortex0, coord, 1.0).rgb;
        } else {
            scene_hdr = texture_sharp(colortex0, coord, 0.66).rgb;
        }
        #else
        if (MC_RENDER_QUALITY > 0.9) {
            scene_hdr  = stex(colortex0).rgb;
        } else {
            scene_hdr  = textureBicubic(colortex0, coord).rgb;
        }
    #endif

    #ifdef bloom_enabled
        vec2 cres       = max(viewSize, vec2(1920.0, 1080.0));

        float bloom_int = 0.014 + max(stex(colortex4).a, 1.0)*0.0025;

        #if dim == -1
            bloom_int  *= 1.5;
        #elif dim == 1
            bloom_int  *= 1.5;
        #endif

            scene_hdr  += texture(colortex3, coord/cres*vec2(1920.0, 1080.0)*0.5).rgb * bloom_int * bloom_intensity;  //apply bloom
    #endif

    #ifdef do_colorgrading
        scene_hdr   = rgb_luma(scene_hdr);
    #endif

    #ifdef manual_exposure_enabled
        scene_hdr  *= rcp(manual_exposure);
    #else
        scene_hdr  *= stex(colortex4).a;
    #endif

    vec3 scene_sdr  = tonemap_hable(scene_hdr);

        scene_sdr   = vibrance_saturation(scene_sdr);

    #ifdef do_colorgrading
        scene_sdr   = gammacurve(scene_sdr);
        scene_sdr   = brightness_contrast(scene_sdr);
    #endif

    #ifdef vignette_enabled
        scene_sdr   = vignette(scene_sdr);
    #endif

        scene_sdr   = dither_image(scene_sdr);

    gl_FragColor    = vec4(scene_sdr, 1.0);
}